require(tidyverse)
require(stringr)
require(stringdist)
library(parallel)


#  ===========================
#' Parameters
#' =========================== 
# TARGET_TEMPLATE <<- c("CCC","VV","CCC","VV","CCC","VV","CCC","VV","CCC")
STEM_TEMPLATE <<- c("CC","VV","CC","VV","CC","VV")
ENDING_TEMPLATE <<- c("CC","VV","CC")
TARGET_TEMPLATE <<- c(STEM_TEMPLATE, ENDING_TEMPLATE)
NULLCONS <<- "#"
NULLVOW <<- "%"
EMPTY_SYLLABLE <<- "##%%##"

PN_scheme <- data.frame(
matrix(
	c(
	"sg1", "1000",
	"sg2", "0100",
	"sg3", "0010",
	"pl1", "1001",
	"pl2", "0101",
	"pe4", "0011"
	), ncol=2, byrow=TRUE)
)
colnames(PN_scheme) = c("PN", "PN_code")



#  ===========================
#' Functions
#' =========================== 

#' IPA and ASCII
#' ---------------------------
parse_ASCII <- function(s){
	unlist(str_extract_all(s, '((\\(.)?.\'?~?\\)?)'))
}

# Includes suprasegmentals in chunks
chunk_ASCII <- function(s){
	unlist(str_extract_all(s, '((\\(.)?.\'?:?~?\\)?)'))
}

ipa2ascii <- function(s){
	for(i in nrow(mappings):1){
		s <- gsub(mappings$IPA[i],mappings$ASCII[i],s, fixed=TRUE)
	}
	s
}

ascii2ipa <- function(s){
	# chunks <- unlist(str_split(s,''))
	slist <- parse_ASCII(s)
	paste(mappings$IPA[match(slist, mappings$ASCII)], collapse="")
}

count_phonemes_ASCII <- function(s){
	# Remove suprasegmentals
	s2 <- remove_suprasegmentals(s)
	slist <- parse_ASCII(s2)
	length(slist)
}

count_phonemes_IPA <- function(s){
	t <- ipa2ascii(s)
	# Remove suprasegmentals
	t2 <- remove_suprasegmentals(t)
	slist <- parse_ASCII(t2)
	length(slist)
}

chunk_phonemes <- function(ipa){
	ipa2ascii(ipa) -> .
	chunk_ASCII(.) -> .
	unlist(lapply(.,ascii2ipa))
}

remove_suprasegmentals <- function(ascii){
	str_replace_all(ascii, ':', '')
}

ascii2vc <- function(s){
	s2 <- remove_suprasegmentals(s)
	slist <- parse_ASCII(s2)
	# as.character(mappings$VC[match(slist, mappings$ASCII)])
	paste(mappings$VC[match(slist, mappings$ASCII)], collapse="")
}

str_rev <- function(x) {
  sapply( x, function(xx) { 
    intToUtf8( rev( utf8ToInt( xx ) ) )
  } )
}


#' Templates
#' ---------------------------
str_tag <- function(x){
	ascii2vc(x)
}

get_slotsizes <- function(template){
	unlist(str_split(paste(template, collapse=""),"")) %>% str_replace_all("C", as.character(NUM_CONS_UNITS)) %>% str_replace_all("V", as.character(NUM_VOW_UNITS)) %>% as.numeric()
	}


extract_ending <- function(ascii){
	phonchunks <- chunk_ASCII(ascii)
	endingChunks <- phonchunks[(NUM_STEM_PHONEMES+1):NUM_TARGET_PHONEMES]
	paste(endingChunks, collapse="")
}

extract_stem <- function(ascii){
	phonchunks <- chunk_ASCII(ascii)
	stemChunks <- phonchunks[1:NUM_STEM_PHONEMES]
	paste(stemChunks, collapse="")
}

template2ipa <- function(s){
	str_replace_all(ascii2ipa(s), "NA", ".")
}

template2ipa2 <- function(s){
	str_replace_all(ascii2ipa(s), "NA", "")
}



match_template <- function(x, template, slotalign="r", patternalign="l"){
	xlist <- chunk_ASCII(x)
	tags <- str_tag(x)
	#
	if(patternalign=="r"){
		template <- rev(template)
		xlist <- rev(xlist)
		tags <- str_rev(tags)
	}

	pattern <- NULL

	for(slots in template){
		type <- substr(slots, 1, 1)
		numSlots <- str_length(slots)
		NULLSYMB <- "#"
		if(type=="C") NULLSYMB <- NULLCONS
		if(type=="V") NULLSYMB <- NULLVOW
		#
		if(length(xlist)==0) {
			filledSlots <- rep(NULLSYMB, numSlots)
			pattern <- c(pattern, paste(filledSlots, collapse=""))
			next
		}
		#
		if(str_sub(tags,1,1)!=type){
			filledSlots <- rep(NULLSYMB, numSlots)
			pattern <- c(pattern, paste(filledSlots, collapse=""))
			next
		}
		#
		len <- str_length(as.vector(str_match(tags, str_c(type,"+"))))
		# Fillers cannot be longer than slots
		# If they are, fill slots and move on
		len <- min(len, numSlots)
		fillers <- xlist[1:len]
		if(patternalign=="r") fillers <- rev(fillers)
		# left-justified:
		if(slotalign=="l") filledSlots <- c(fillers, rep(NULLSYMB, numSlots - len))
		# right-justified:
		if(slotalign=="r") filledSlots <- c(rep(NULLSYMB, numSlots - len), fillers)
		# Append to pattern
		pattern <- c(pattern, paste(filledSlots, collapse=""))
		#
		tags <- str_sub(tags, len+1)
		xlist <- xlist[-(1:len)]
	}
	#
	if(length(xlist)!=0 & !is.na(x)) {
			print(str_c("WARNING: String ", x," too long for template!"))
		}
	#
	if(patternalign=="r") pattern <- rev(pattern)
	paste(pattern, collapse="")
}



#' Encoding
#' ---------------------------
encode <- function(s){
	# chunks <- unlist(str_split(s,''))
	chunks <- chunk_ASCII(s)
	code <- paste(ipabin$code[match(chunks, ipabin$ASCII)], collapse="")
	str_c('s',code)
}


phonchunk_binary <- function(bin, template=TARGET_TEMPLATE){
	clustersizes <- nchar(template)
	slotsizes <- unlist(str_split(paste(template, collapse=""),""))
	slotsizes <- str_replace_all(slotsizes, "C", as.character(NUM_CONS_UNITS))
	slotsizes <- str_replace_all(slotsizes, "V", as.character(NUM_VOW_UNITS))
	slotsizes <- as.numeric(slotsizes)
	binlist <- unlist(str_split(bin,""))
	groupings <- rep(1:sum(clustersizes), slotsizes)
	chunks <- split(binlist, groupings)
	chunks <- as.vector(unlist(lapply(chunks, function(x) paste(x, collapse=""))))
}


phonchunk_activations <- function(acts, template=TARGET_TEMPLATE){
	# slotsizes <- unlist(str_split(paste(template, collapse=""),""))
	# slotsizes <- str_replace_all(slotsizes, "C", NUM_CONS_UNITS)
	# slotsizes <- str_replace_all(slotsizes, "V", NUM_VOW_UNITS)
	# slotsizes <- as.numeric(slotsizes)
	slotsizes <- get_slotsizes(template)
	nUnits <- sum(slotsizes)
	nPhonems <- sum(nchar(template))
	nTrials <- length(acts)/nUnits
	n_total_phonemes <- nPhonems*nTrials
	groupings <- rep(1:n_total_phonemes, rep(slotsizes, n_total_phonemes/nPhonems))
	chunks <- split(acts, groupings)
	# chunks <- as.vector(unlist(lapply(chunks, function(x) paste(x, collapse=""))))
}


decode <- function(code, template=TARGET_TEMPLATE){
	code <- str_replace(code, 's','')
	chunks <- phonchunk_binary(code, template)
	paste(ipabin$ASCII[match(chunks, ipabin$code)], collapse="")
}


decode_phoneme <- function(chunks){
	ipabin$ASCII[match(chunks, ipabin$code)]
}

# decode_nearest2 <- function(acts, template=TARGET_TEMPLATE){
# 	phonchunks <- phonchunk_activations(acts, template)
# 	lapply(phonchunks,decode)
# 	# vapply(phonchunks,print, "a")
# 	# split(acts, ceiling(seq_along(acts)/8))
# }

decode_activations <- function(acts, template=TARGET_TEMPLATE){
	phonchunks <- phonchunk_activations(acts, template)
	phonchunks <- as.vector(unlist(lapply(phonchunks, function(x) paste(x, collapse=""))))
	decode_phoneme(phonchunks)
	# ipabin$ASCII[match(phonchunks, ipabin$code)]
	# vapply(phonchunks,decode, "a")
	# vapply(phonchunks,print, "a")
	# split(acts, ceiling(seq_along(acts)/8))
}

decode_activations_nearest <- function(acts, template=TARGET_TEMPLATE){
	phonchunks <- phonchunk_activations(acts, template)
	# vapply(phonchunks,decode_nearest_phoneme, "a")
	as.vector(sapply(phonchunks, decode_nearest_phoneme))
	# vapply(phonchunks,print, "a")
	# split(acts, ceiling(seq_along(acts)/8))
}


decode_activations_nearest_parallel <- function(acts, template=TARGET_TEMPLATE){
	phonchunks <- phonchunk_activations(acts, template)
	# phonchunks <- as.vector(unlist(lapply(phonchunks, function(x) paste(x, collapse=""))))
	## stringdist(ipabin$code, "1001101")
	## ipabin$ASCII[stringdist(chunks, ipabin$code)]
	cluster <- makeCluster(4)
	clusterExport(cluster, c("ipabin"))
	## registerDoParallel(cluster) 
	## startt <- proc.time()
	## as.vector(vapply(phonchunks, decode_nearest_phoneme2, "a"))
	y <- as.vector(parSapply(cluster, phonchunks, decode_nearest_phoneme))
	## proc.time() - startt
	stopCluster(cluster) # Stop cluster
	## vapply(phonchunks,print, "a")
	## split(acts, ceiling(seq_along(acts)/8))
	y
}

decode_activations_nearest2 <- function(acts, template=TARGET_TEMPLATE){
	phonchunks <- phonchunk_activations(acts, template)
	phonchunks <- as.vector(unlist(lapply(phonchunks, function(x) paste(x, collapse=""))))
	## stringdist(ipabin$code, "1001101")
	## ipabin$ASCII[stringdist(chunks, ipabin$code)]
	as.vector(sapply(phonchunks, decode_nearest_phoneme2))
	## vapply(phonchunks,print, "a")
	## split(acts, ceiling(seq_along(acts)/8))
}


decode_activations_nearest2_parallel <- function(acts, template=TARGET_TEMPLATE){
	phonchunks <- phonchunk_activations(acts, template)
	# phonchunks <- as.vector(unlist(lapply(phonchunks, function(x) paste(x, collapse=""))))
	## stringdist(ipabin$code, "1001101")
	## ipabin$ASCII[stringdist(chunks, ipabin$code)]
	cluster <- makeCluster(4)
	clusterExport(cluster, c("stringdist", "ipabin"))
	## registerDoParallel(cluster) 
	## startt <- proc.time()
	## as.vector(vapply(phonchunks, decode_nearest_phoneme2, "a"))
	y <- as.vector(parSapply(cluster, phonchunks, decode_nearest_phoneme2))
	## proc.time() - startt
	stopCluster(cluster) # Stop cluster
	## vapply(phonchunks,print, "a")
	## split(acts, ceiling(seq_along(acts)/8))
	y
}



# pcode = c(0.003262730,0.004449270,0.005037040,0.000457661,0.000702493,0.000752842,0.000752842)

decode_nearest_phoneme <- function(pcode){
	#' filter for consonants/vowels
	targettable <- subset(ipabin, nchar(code)==length(pcode))
	# ptargets <- separate(targettable, code, paste("U", 1:length(pcode), sep=""), sep=1:(length(pcode)-1), remove=FALSE) %>% select(matches("U\\d"))
	# ptargets <- apply(ptargets, 2, as.numeric)
	ptargets <- matrix(as.numeric(unlist(strsplit(targettable$code, ""))), ncol=length(pcode), byrow=TRUE)
	# ptargets <- mappings[,4:ncol(mappings)]
	# dm <- as.matrix(dist(rbind(unlist(pcode),ptargets)))	
	# pdist <- as.vector(dm[-1,1])
	pdist <- sqrt(colSums((unlist(pcode)-t(ptargets))^2))
	phoneme <- as.vector(targettable$ASCII)[which(pdist==min(pdist))[1]]
	phoneme
}

decode_nearest_phoneme2 <- function(x){
	y <- ipabin$ASCII[match(x, ipabin$code)]
	if(is.na(y)) {
		pdist <- stringdist(ipabin$code, x)
		y <- as.vector(ipabin$ASCII)[which(pdist==min(pdist))[1]]
	}
	y	
}



#' Localist lemma encoding
#' ---------------------------
encode_localist_input <- function(x){
	#' Assign random numbers to lemmas
	lemmas <- unique(x$Lemma_Class)
	n <- length(lemmas)
	set.seed(4711)
	units <- sample(1:length(lemmas))
	x <- tbl_df(data.frame(Lemma_Class=lemmas, LemmaCode=units)) %>% right_join(x)
	x$Input <- sapply(x$LemmaCode, function(y) paste(c("s", rep(0,y-1), 1, rep(0,n-y)), collapse=""))
	#' Add target codes to input
	x <- unite(x, Input, Input, PN_code, sep="")
	#
	x
}



#' Write examples
#' ---------------------------

#' Writes example files with non-CDS tokens at tokencount 0.  
#' (So they won't be included in the probabilistic training set.)
writeExamples <- function(examples, filename){
	input1 <- str_trim(gsub('', ' ', sub('s','', examples$Input)))
	target1 <- str_trim(gsub('', ' ', sub('s','', examples$Target)))
	vnames <- str_c("name: {", examples$Label,"}")
	freqs <- paste("freq:", examples$freq)
	input <- paste("I:", input1)
	target <- paste("T:", target1, ";")
	out <- c(rbind(vnames,freqs,input,target))
	write(out, filename)
}


# #' Writes example files with non-CDS tokens at tokencount 1.  
# #' (So they have a chance to appear in the probabilistic training set.)
# writeExamplesAllTokens <- function(examples, filename){
# 	input1 <- str_trim(gsub('', ' ', sub('s','', examples$Input)))
# 	target1 <- str_trim(gsub('', ' ', sub('s','', examples$Target)))
# 	vnames <- str_c("name: {", examples$Label,"}")
# 	freqs <- paste("freq:", examples$Token_count+1)
# 	input <- paste("I:", input1)
# 	target <- paste("T:", target1, ";")
# 	out <- c(rbind(vnames,freqs,input,target))
# 	write(out, filename)
# }


# bits <- 108
# paste(rep("Stem", bits))







#  ===========================
#' IPA-ASCII mappings
#' =========================== 
mappings <<- read_csv("SAMPA.csv")



#  ===========================
#' IPA-binary mappings
#' =========================== 
ipabin <<- read_csv("IPA-binary.csv")

#' Load transcription rules for current language and create phoneme inventory
rules <- read_csv("transcription-rules.csv")
inventory <- rules$IPA %>% chunk_phonemes() %>% unique()

#' ASCII
ipabin <<- mutate(ipabin, ASCII = ipa2ascii(IPA))

#' Codes
ipabin <<- ipabin %>% filter(IPA!="ː") %>% unite(code, contains("Bin_"), sep="") 
# ipabin <<- ipabin %>% unite(code, contains("Bin_"), sep="") 
ipabin$code <- str_replace_all(ipabin$code, "NA", "")

#' Add separate lengthening symbol
# ipabin$IPA[ipabin$IPA=="ː"] <- "="
# ipabin$ASCII[ipabin$IPA=="="] <- "="
# ipabin$code[ipabin$IPA=="="] <- str_replace(ipabin$code[ipabin$IPA=="%"], "(0)$", "1")
# ipabin$consonant[ipabin$IPA=="="] <- 0
# ipabin$IPA[ipabin$IPA=="="] <- "="
# mappings <<- rbind(mappings, c(NA, "=", NA, "=", "V", NA, NA, NA, NA))



ipabin[is.na(ipabin)] <- 0

#' Add lengthened phonemes
longphons <- filter(ipabin, IPA!="#" & IPA!="%" & IPA!="=" & nchar(IPA)<3) %>% mutate(IPA = str_c(IPA,"ː"), ASCII = str_c(ASCII, ":"), code = str_replace(code, "(0)$", "1"))
ipabin <<- bind_rows(ipabin, longphons)

# print(ipabin, n=100)
# tail(ipabin)

#' Eliminate unused phonemes
# inventory <- c(inventory, "=")
ipabin <- filter(ipabin, IPA%in%inventory | IPA%in%c(NULLCONS, NULLVOW))

#' Catch duplicates
dups <- filter(ipabin, duplicated(code))$code
# filter(ipabin, code%in%dups) %>% print(n=100)



#  ===========================
#' Set global variables
#' =========================== 
NUM_CONS_UNITS <<- ipabin %>% filter(consonant==1) %>% with(nchar(code[1]))
NUM_VOW_UNITS <<- ipabin %>% filter(consonant==0) %>% with(nchar(code[1]))
NUM_TARGET_UNITS <<- sum(get_slotsizes(TARGET_TEMPLATE))
NUM_STEM_UNITS <<- sum(get_slotsizes(STEM_TEMPLATE))
NUM_END_UNITS <<- NUM_TARGET_UNITS - NUM_STEM_UNITS	
NUM_TARGET_PHONEMES <<- sum(nchar(TARGET_TEMPLATE))
NUM_STEM_PHONEMES <<- sum(nchar(STEM_TEMPLATE))
NUM_END_PHONEMES <<- NUM_TARGET_PHONEMES - NUM_STEM_PHONEMES

