
require(tidyverse, quietly=TRUE, warn.conflicts=FALSE)
require(stringr, quietly=TRUE, warn.conflicts=FALSE)

source("fct.extract_activations.R")
source("fct.compute_errors.R")
source("fct.encode.R")


args = commandArgs(trailingOnly=TRUE)



process <- function(filename, actsfile, resfile, summfile){
	acts <- extract_activations(filename)
	errors <- compute_errors(acts)

	ez <- gzfile(resfile, open="ab")
	if(file.size(resfile)==0) {
		write.table(errors, ez, col.names=TRUE, row.names=FALSE, append=FALSE)
	} else {
		write.table(errors, ez, col.names=FALSE, row.names=FALSE, append=TRUE)
	}
	close(ez)

	summ <- summarise_each(errors[,-c(1,3:4)], funs(mean))

	if(!file.exists(summfile)) {
		write.table(summ, summfile, append=FALSE, col.names=TRUE, row.names=FALSE)
	} else {
		write.table(summ, summfile, append=TRUE, col.names=FALSE, row.names=FALSE)
	}
}



summarise_net <- function(netname, path){
	train <- read.table(str_c(path,"/",netname,".train.summary.txt"), header=TRUE)
	test <- read.table(str_c(path,"/",netname,".test.summary.txt"), header=TRUE)
	novel <- read.table(str_c(path,"/",netname,".novel.summary.txt"), header=TRUE)
	test$Set <- "test"
	train$Set <- "train"
	novel$Set <- "novel"

	dat <- bind_rows(test,train,novel)
	dat <- gather(dat, Measure, Value, c(PhonError, Error, PhonAcc))

	(p <- ggplot(filter(dat, Epoch>10 & Measure%in%c("PhonAcc", "Error")), aes(Epoch, Value, color=Set, shape=Set, group=Set)) +
	# geom_line(stat='summary', fun.data="mean_se") +
	# geom_point(stat='summary') +
  # stat_summary(fun.data = "mean_se") +
		geom_line() +
		geom_point() +
		scale_x_log10() +
		ylim(0,1) +
		facet_grid(Measure~., scales='free')
		)

	print(str_c(path,"/summary.",netname,".pdf"))
	ggsave(p, file=str_c(path,"/summary.",netname,".pdf"))
}



process_files <- function(path, maxNetworks, maxEpoch){
	files <- list.files(path=path, pattern=".*\\.\\d+\\.out.gz")
	files <- files[as.numeric(str_match(files, '\\.(\\d+)\\.out')[,2])<=maxEpoch]
	files <- files[as.numeric(str_match(files, '^[^.\\d]+(\\d\\d?)\\.')[,2])<=maxNetworks]
	unlink(str_c(path,'/*.results.txt.gz'))
	unlink(str_c(path,'/*.summary.txt'))
	for(f in files){
		filename <- paste(path,f,sep='/')
		netname <- str_match(f, "^([^.]+)\\.")[2]
		actsfile <- str_replace(filename, 'out.gz', 'acts.txt.gz')
		resfile <- str_replace(filename, '\\.\\d+\\.out.gz', '.results.txt.gz')
		summfile <- str_replace(filename, '\\.\\d+\\.out.gz', '.summary.txt')
		process(filename, actsfile, resfile, summfile)
		# source("script.summarise.R")
	}
}


summarise_results <- function(path){
	files <- list.files(path=path, pattern=".*\\.\\d+\\.out.gz")
	netnames <- unique(str_match(files, "^([^.]+)\\.")[,2])
	for(n in netnames){
		netname <- n
		# path <- prefix
		# try(source("script.summarise.R"))
		try(summarise_net(n, path))
	}
}




if(length(args)>0){
	path <- args[1]
	if(length(args)>1) maxNetworks <- as.numeric(args[2]) else maxNetworks <- 100
	if(length(args)>2) maxEpoch <- as.numeric(args[3]) else maxEpoch <- 10e6
	start <- proc.time()
	process_files(path, maxNetworks, maxEpoch)
	summarise_results(path)
	print(proc.time() - start) # End clock
}
