
require(tidyverse)
options(dplyr.width = Inf)
require(stringr)
require(xtable)

source("fct.encode.R")
# source("script.process.R")
# source("fct.compute_errors.R")

args = commandArgs(trailingOnly=TRUE)


read_net <- function(itemsFile, resultsFile){
	items <- tbl_df(read.csv(itemsFile, header=TRUE, encoding="UTF-8", stringsAsFactors=FALSE)) 
	items <- items %>% select(ID) %>%
		mutate(Example=1:nrow(items))
	#
	results <- tbl_df(read.table(resultsFile, header=TRUE, stringsAsFactors=FALSE, encoding="UTF-8"))
	results <- left_join(items, results, by="Example")
	results
}


reprocess_files <- function(prefix, path='output'){
	files <- list.files(path=path, pattern=str_c(prefix,'.*\\.\\d+\\.out.gz'))
	unlink(str_c(path,'/',prefix,'*.results.txt.gz'))
	unlink(str_c(path,'/',prefix,'*.summary.txt'))
	for(f in files){
		filename <- paste(path,f,sep='/')
		actsfile <- str_replace(filename, 'out.gz', 'acts.txt.gz')
		resfile <- str_replace(filename, '\\.\\d+\\.out.gz', '.results.txt.gz')
		summfile <- str_replace(filename, '\\.\\d+\\.out.gz', '.summary.txt')
		process(filename, actsfile, resfile, summfile)
	}
}

read_test <- function(netname, path="output"){
	itemsFile <- str_c(path,'/',netname,'.testset.csv')
	resultsFile <- str_c(path,'/',netname,'.test.results.txt.gz')
	results <- read_net(itemsFile, resultsFile)
	results$net <- netname
	results
}

read_results <- function(setname, netname, path='output'){
	itemsFile <- str_c(path,'/',netname,'.',setname,'set.csv')
	resultsFile <- str_c(path,'/',netname,'.',setname,'.results.txt.gz')
	results <- read_net(itemsFile, resultsFile)
	results$net <- netname
	results
}


# path = "modelH200FIVerbsProdSimplvow"

prepare_data <- function(path){

# ===========================
# Load data 
# ===========================
	files <- list.files(path=path, pattern=".+\\..+\\.results.txt.gz")
	netnames <- str_match(files, "^([^.]+)\\.")[,2] %>% unique()

	testres <- lapply(netnames, function(x){read_results("test", x, path)}) %>% do.call("bind_rows", .)
	trainres <- lapply(netnames, function(x){read_results("train", x, path)}) %>% do.call("bind_rows", .)
	novelres <- lapply(netnames, function(x){read_results("novel", x, path)}) %>% do.call("bind_rows", .)

	trainset <- read_csv("trainset-all.encoded.csv")
	trainset <- select(trainset, -Input, -Target)

	# PN Labels
	# ---------------------------
	trainset <- mutate(trainset, Person = str_match(PN, "..(\\d)")[,2])
	trainset <- mutate(trainset, Person = ifelse(Person==4, 1, Person))
	trainset <- mutate(trainset, Number = str_match(PN, "(..)\\d")[,2])

	trainset <- mutate(trainset, PN=factor(PN), Person=factor(Person), Number=factor(Number), Subclass=factor(Subclass))
	# trainres$PN <- factor(trainres$PN)
	# testres$PN <- factor(testres$PN)



	# Trained vs. untrained
	# ---------------------------
	#' Train set:
	with(trainset, table(Subclass, PN))

	#' Items excluded from train set by net:
	trainsetExcluded <- setdiff(distinct(select(novelres, ID, net)), distinct(select(trainres, ID, net)))

	with(trainsetExcluded, table(net))

	# trainsetExcluded$trained <- FALSE
	novelres$trained <- FALSE

	# # Which Lemmas belong to the set of withhold items:
	# 	holdLemmas <- testres %>% filter(!trained) %>% with(ID) %>% unique()
	# 	length(holdLemmas)
	# 	trainres <- mutate(trainres, HoldSet=ifelse(Lemma%in%holdLemmas, TRUE, FALSE))
	# 	testres <- mutate(testres, HoldSet=ifelse(Lemma%in%holdLemmas, TRUE, FALSE))

	# testres %>% select(ID, Lemma, Token, net, trained) %>% distinct() %>% with(table(net,trained))


	# Frequency
	# ---------------------------
	# trainset$Lemma_freq <- trainset$Lemma_freq * 794000
	# Freq grouping
	# trainset <- mutate(trainset, Token_freq_group = factor(cut2(Token_freq, g=3), labels=c('low','mid','high')))


	#' Join test sets
	#' ---------------------------
	testres$Testset <- "test"
	trainres$Testset <- "train"
	novelres$Testset <- "novel"

	simdata <- bind_rows(trainres, testres, novelres) %>% mutate(trained=ifelse(is.na(trained), TRUE, trained))


	# Repair ID
	# ---------------------------
	if(grepl('[^-]+-[^-]+-[^-]+-[^-]+', simdata$ID[1])){
		simdata <- separate(simdata, ID, c("Class", "Group", "Lemma", "PN"), sep="-", remove=TRUE)
		simdata <- unite(simdata, ID2, Lemma, PN, sep="-", remove=TRUE) %>% select(-Class, -Group)

		trainset <- unite(trainset, ID2, Lemma, PN, sep="-", remove=FALSE)
		simdata <- left_join(simdata, select(trainset, ID, ID2))

		simdata <- select(simdata, -ID2)
		trainset <- select(trainset, -ID2)
	}

	
	simdata <- simdata %>% separate(ID, c("Lemma", "PN", "Token"), sep="-", remove=FALSE)
	simdata <- left_join(simdata, select(trainset, ID, Class, Subclass))

	# Accuracy
	# ---------------------------
	simdata <- simdata %>% mutate(Acc = ifelse(PhonError==0, 1, 0))


	# Suffix / stem errors
	# ---------------------------
	simdata <- rowwise(simdata) %>% mutate(
		OutputString_Stem = extract_stem(OutputString),
		OutputString_Ending = extract_ending(OutputString),
		TargetString_Stem = extract_stem(TargetString),
		TargetString_Ending = extract_ending(TargetString)
		) %>% ungroup()

	#' Correct for endings that are only a lengthened vowel on the stem
	filter(trainset, is.na(Ending_ASCII) & grepl(":$", TargetStem_ASCII) & !grepl(":$", Stem_ASCII))
	longVowEndings <- filter(trainset, is.na(Ending_ASCII) & grepl(":$", TargetStem_ASCII) & !grepl(":$", Stem_ASCII)) %>% with(ID)
	# trainset <- mutate(trainset, TargetString_Ending = ifelse(TargetString_Ending==EMPTY_SYLLABLE & grepl(":%?$", TargetString_Stem) & !grepl(":%?$", Stem_ASCII), ":", TargetString_Ending))
	trainset <- mutate(trainset, TargetString_Ending = ifelse(ID%in%longVowEndings, ":", TargetString_Ending))
	filter(simdata, ID%in%longVowEndings)
	simdata <- mutate(simdata, TargetString_Ending = ifelse(ID%in%longVowEndings, ":", TargetString_Ending))
	simdata <- mutate(simdata, OutputString_Ending = ifelse(OutputString_Ending==EMPTY_SYLLABLE & grepl(":%?$", OutputString_Stem) & ID%in%longVowEndings, ":", OutputString_Ending))


	#' Stem/suffix 
	simdata <- simdata %>% mutate(StemError = ifelse(OutputString_Stem==TargetString_Stem, 0, 1), SuffixError = ifelse(OutputString_Ending==TargetString_Ending, 0, 1))

	simdata <- simdata %>% mutate(StemOnlyError = ifelse(StemError & !SuffixError, 1, 0), SuffixOnlyError = ifelse(SuffixError & !StemError, 1, 0), BothError = ifelse(StemError & SuffixError, 1, 0), Correct = ifelse(!StemError & !SuffixError, 1, 0))

	simdata <- simdata %>% mutate(StemAcc = 1-StemError, SuffixAcc = 1-SuffixError)



	#' Error Scores
	#' --------------------  

	#' Possible forms
	possibleForms <- unique(simdata$TargetString)
	possibleStems <- unique(simdata$TargetString_Stem)
	possibleEndings <- unique(simdata$TargetString_Ending)
	simdata <- mutate(simdata, possible=ifelse(OutputString%in%possibleForms, 1, 0), possible_Stem=ifelse(OutputString_Stem%in%possibleStems, 1, 0), possible_Ending=ifelse(OutputString_Ending%in%possibleEndings, 1, 0))

	#' Possible suffixes
	suffixes <- trainset %>% select(TargetString_Ending, PN, Class, Subclass) %>% distinct()
	# suffixes2 <- filter(suffixes, TargetString_Ending != EMPTY_SYLLABLE)
	suffixesPN <- suffixes %>% select(TargetString_Ending, PN) %>% distinct()
	# Are there any duplicated endings?
	duplicated(suffixesPN$TargetString_Ending)
	suffixesClass <- suffixes %>% select(TargetString_Ending, Class) %>% distinct()
	suffixesSubclass <- suffixes %>% select(TargetString_Ending, Subclass) %>% distinct()

	dim(simdata)
	simdata <- left_join(simdata, rename(suffixesPN, OutputString_Ending=TargetString_Ending, OutputPN=PN))
	dim(simdata)

	simdata <- mutate(simdata, PNerror = ifelse(SuffixError & OutputPN!=PN, 1, 0))
	simdata <- mutate(simdata, PNerror = ifelse(is.na(OutputPN), 1, 0))
	simdata <- mutate(simdata, PNerror = ifelse(!SuffixError, 0, PNerror))
	
	
	#' Define different scores  
	simdata <- mutate(simdata, unscorable = ifelse(!possible_Ending, TRUE, FALSE))
	#' Responses were scored as correct (1), incorrect (0) or NA under three different coding columns:  
	#' i. PN affix: whether or not the child produced an affix which belongs to the correct PN.
	simdata <- mutate(simdata, Acc_PN = ifelse(!SuffixError | !PNerror, 1, 0))
	# simdata <- mutate(simdata, Acc_PN = ifelse(unscorable, NA, Acc_PN))
	#' ii. Stem: whether or not the stem was produced correctly (NA if PN affix is incorrect)
	simdata <- mutate(simdata, Acc_Stem = ifelse(!StemError, 1, 0))
	simdata <- mutate(simdata, Acc_Stem = ifelse(!Acc_PN, NA, Acc_Stem))

	#' iii. Class affix: whether or not the produced affix was correct for that class (NA if case affix is incorrect)
	simdata <- mutate(rowwise(simdata), Acc_Class = ifelse(OutputString_Ending%in%suffixesClass$TargetString_Ending[suffixesClass$Class==Class], 1, 0)) %>% ungroup()
	simdata <- mutate(simdata, Acc_Class = ifelse(!Acc_PN, NA, Acc_Class))
	simdata <- mutate(simdata, Acc_Class = ifelse(is.na(Acc_PN), NA, Acc_Class))

	#' iv. Total suffix
	simdata <- mutate(simdata, Acc_Suffix = ifelse(!SuffixError, 1, 0))
	# filter(simdata, Acc_Class==1)
	# filter(simdata, Acc_Stem==1)
	# filter(simdata, Acc_Class==1 & Acc_PN==1)
	# filter(simdata, is.na(Acc_PN))


	# Make factors
	# ---------------------------
	trainset <- mutate(trainset, Class=factor(Class), PN=factor(PN), Subclass=factor(Subclass))

	# # Epoch labels
	# # ---------------------------
	# simdata <- mutate(simdata, EpochNum = Epoch, EpochFctr = factor(Epoch))

	trainset <<- trainset
	simdata <<- simdata

}




if(length(args)>0){
	start <- proc.time()
	# 
	path <- args[1]
	simdata <<- NULL
	trainset <<- NULL
	prepare_data(path)
	save(simdata, trainset, file=str_c(path,".data.Rd"))
	# 
	print(proc.time() - start)
}


